#ifndef LIB_DBG_DEBUG_MQLAPI_TRACER_MQH_INCLUDED
#define LIB_DBG_DEBUG_MQLAPI_TRACER_MQH_INCLUDED
#property version "5.50"
/**********************************************************************************
 * Copyright (C) 2020 Dominik Egert <info@freie-netze.de>
 *
 * This file is the MQL-API base tracer object include file.
 *
 * Lisence applied: GPLv2
 *                  https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 *
 * Author Dominik Egert / Freie Netze UG.
 **********************************************************************************
 *
 *  Version: 4120
 *  State: public
 *
 *  File information
 *  ================
 *
*/




//*********************************************************************************************************************************************************/
//  BEGIN MQLAPI tracer
//


/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// MQL API general function scheme
//
//
//      This function wrapping scheme is used for all mql-api-functions.
//
//      In basic there are three types of function wrapping methods to be used.
//          - Functions of VOID-Type
//          - Functions of RETURN-Type
//          - Variable input parameter functions
//
//      Representation of the original function signature as documented was most important, to prevent differences
//      by the wrapper and the original function. - In case of doubt, the more reliable option was chosen.
//
//
//  Example function wrapper for RETURN-Type:
//
//      Original function signature from documentation:
//
//          int  WebRequest(
//             const string      method,           // HTTP method
//             const string      url,              // URL
//             const string      headers,          // headers
//             int               timeout,          // timeout
//             const char        &data[],          // the array of the HTTP message body
//             char              &result[],        // an array containing server response data
//             string            &result_headers   // headers of server response
//             );
//
//      Wrapper replacement:
//
//          int dbg_mqlapi_trace_func_WebRequest(const string method, const string url, const string headers, int timeout, const char& data[], char& result[], string& result_headers)
//          {
//              // Trace call setup
//              DBG_MQLAPI_TRACE_FUNCTION_CALL_SETUP;
//
//              // All input variables are stringified to "_params_in"-member variable.
//              DBG_MQLAPI_TRACE_FUNC_PARAMS_STR(7, method, url, headers, timeout, data, result, result_headers);
//
//              // The actual wrapped MQL api function call
//              int retval = WebRequest(method, url, headers, timeout, data, result, result_headers);
//
//              // In case output is provided by parameter references, output-print is avialable in "_params_out" member variable.
//              DBG_MQLAPI_TRACE_FUNC_PARAM_RESULTS_STR(2, result, result_headers, NULL, NULL, NULL, NULL, NULL);
//
//              // Output return value given by mql function
//              DBG_MQLAPI_TRACE_PRINT_RETURN(retval);
//
//              // Finalize call tracing
//              DBG_MQLAPI_TRACE_FUNCTION_CALL_FINALIZE;
//              return(retval);
//          };
//
//
//  Example function wrapper for VOID-Type:
//
//      Original function signature form documentation:
//
//          void  ArrayPrint(
//             const void&   array[],             // printed array
//             uint          digits=_Digits,      // number of decimal places
//             const string  separator=NULL,      // separator of the structure field values
//             ulong         start=0,             // first printed element index
//             ulong         count=WHOLE_ARRAY,   // number of printed elements
//             ulong         flags=ARRAYPRINT_HEADER|ARRAYPRINT_INDEX|ARRAYPRINT_LIMIT|ARRAYPRINT_ALIGN
//             );
//
//      Wrapper replacement:
//
//          template <typename T>
//          void dbg_mqlapi_trace_ArrayPrint(const T& arr[], uint digits = UINT_MAX, const string separator = NULL, ulong start = 0, ulong count = WHOLE_ARRAY, ulong flags = ARRAYPRINT_HEADER|ARRAYPRINT_INDEX|ARRAYPRINT_LIMIT|ARRAYPRINT_ALIGN)
//          {
//              // Trace call setup
//              DBG_MQLAPI_TRACE_FUNCTION_CALL_SETUP;
//
//              // Additional setup to ensure function compatibility to the original signature.
//              int _digits = (digits == UINT_MAX) ? _Digits : digits;
//
//              // All input variables are stringified to "_params_in"-member variable.
//              DBG_MQLAPI_TRACE_FUNC_PARAMS_STR(6, arr, _digits, separator, start, count, flags);
//
//              // The actual wrapped MQL api function call
//              ArrayPrint(arr, _digits, separator, start, count, flags);
//
//              // Output void type function
//              DBG_MQLAPI_TRACE_PRINT;
//
//              // FInalize call tracing
//              DBG_MQLAPI_TRACE_FUNCTION_CALL_FINALIZE;
//              return;
//          };
//
//
//
/////////////////////////////////////////////////////////////////////////////////////////////////////




/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// MQL API tracer special function definitions
//

    // MQL API Function IDs for variable inputs
    #define DBG_MQL_API_TRACING_FUNCTION_ID_ALERT                                   -1
    #define DBG_MQL_API_TRACING_FUNCTION_ID_COMMENT                                 -2
    #define DBG_MQL_API_TRACING_FUNCTION_ID_PRINT                                   -3
    #define DBG_MQL_API_TRACING_FUNCTION_ID_PRINTF                                  -4
    #define DBG_MQL_API_TRACING_FUNCTION_ID_PRINTFORMAT                             -5
    #define DBG_MQL_API_TRACING_FUNCTION_ID_STRINGFORMAT                            -6
    #define DBG_MQL_API_TRACING_FUNCTION_ID_PRINTFORMAT_ERROR                       -126

    // Rerouting (c/c++) standard functions to MQL-version Math*
    #define DBG_MQL_API_TRACING_FUNCTION_ID_ACOS                                    -7
    #define DBG_MQL_API_TRACING_FUNCTION_ID_ASIN                                    -8
    #define DBG_MQL_API_TRACING_FUNCTION_ID_ATAN                                    -9
    #define DBG_MQL_API_TRACING_FUNCTION_ID_CEIL                                    -10
    #define DBG_MQL_API_TRACING_FUNCTION_ID_COS                                     -11
    #define DBG_MQL_API_TRACING_FUNCTION_ID_EXP                                     -12
    #define DBG_MQL_API_TRACING_FUNCTION_ID_FABS                                    -13
    #define DBG_MQL_API_TRACING_FUNCTION_ID_FLOOR                                   -14
    #define DBG_MQL_API_TRACING_FUNCTION_ID_FMAX                                    -15
    #define DBG_MQL_API_TRACING_FUNCTION_ID_FMIN                                    -16
    #define DBG_MQL_API_TRACING_FUNCTION_ID_FMOD                                    -17
    #define DBG_MQL_API_TRACING_FUNCTION_ID_LOG                                     -18
    #define DBG_MQL_API_TRACING_FUNCTION_ID_LOG10                                   -19
    #define DBG_MQL_API_TRACING_FUNCTION_ID_POW                                     -20
    #define DBG_MQL_API_TRACING_FUNCTION_ID_RAND                                    -21
    #define DBG_MQL_API_TRACING_FUNCTION_ID_ROUND                                   -22
    #define DBG_MQL_API_TRACING_FUNCTION_ID_SIN                                     -23
    #define DBG_MQL_API_TRACING_FUNCTION_ID_SQRT                                    -24
    #define DBG_MQL_API_TRACING_FUNCTION_ID_SRAND                                   -25
    #define DBG_MQL_API_TRACING_FUNCTION_ID_TAN                                     -26


//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// MT4/5 API function call meta object
//
#ifndef __MQL4__
    namespace dbg_lib
    {

#endif


    template <typename EXT_OBJ>    
#ifndef LIB_MQLAPI_TRACE_CUSTOM_EXTENSION
    class LIB_DBG_NAMESPACE_DEF(dbg_lib, mqlapi_call_tracer_object) : public LIB_DBG_NAMESPACE_DEF(dbg_lib, mqlapi_call_tracer_object_data)<EXT_OBJ>

#else
    class LIB_DBG_NAMESPACE_DEF(dbg_lib, mqlapi_call_tracer_object) : public LIB_DBG_NAMESPACE_DEF(dbg_lib, userapi_call_tracer_custom_object)<EXT_OBJ>

#endif
    {
        public:
        // Constructor

            LIB_DBG_NAMESPACE_DEF(dbg_lib, mqlapi_call_tracer_object)()
            {
                ::printf(" >>> LIB_DEBUG: MQLAPI-Tracer enabled <<< \n   - Note: initially clearing _LastError variable now.\n");
                ::ResetLastError();
            };

            LIB_DBG_NAMESPACE_DEF(dbg_lib, mqlapi_call_tracer_object)(const LIB_DBG_NAMESPACE_DEF(dbg_lib, mqlapi_call_tracer_object)& p_in)
            { _global_trace_data[tracer_call_stack_depth] = p_in._global_trace_data[p_in.tracer_call_stack_depth]; };

        // Debug location recorder

            LIB_DBG_NAMESPACE_DEF(dbg_lib, mqlapi_call_tracer_object)* trace_call_details(const int _line_in, const char _func_id, const string _file_in, const string _function_in, const string _mqlapi_func, const string _api_name = NULL)
            {
                // Increase stack call counter
                stack_increment();
                
                // Prestate setup
                _global_trace_data[tracer_call_stack_depth]._params_out         = NULL;
                _global_trace_data[tracer_call_stack_depth]._params_in          = NULL;
                _global_trace_data[tracer_call_stack_depth]._api_name_          = (_global_trace_data[tracer_call_stack_depth]._api_name_ == NULL) ? _api_name : _global_trace_data[tracer_call_stack_depth]._api_name_;
                _global_trace_data[tracer_call_stack_depth]._print_params_      = LIB_MQLAPI_TRACE_SHOW_PARAMS;
                _global_trace_data[tracer_call_stack_depth]._print_result_      = LIB_MQLAPI_TRACE_SHOW_RESULT;
                _global_trace_data[tracer_call_stack_depth]._perf_runtime_      = NULL;
                _global_trace_data[tracer_call_stack_depth]._prev_err_state     = _LastError;

                // Trace setup details
                _global_trace_data[tracer_call_stack_depth]._line               = _line_in;
                _global_trace_data[tracer_call_stack_depth]._rtol_op            = _func_id;
                _global_trace_data[tracer_call_stack_depth]._file               = _file_in;
                _global_trace_data[tracer_call_stack_depth]._function           = _function_in;
                _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname    = _mqlapi_func;

                // Overwrite function name, if applicable
                switch(_global_trace_data[tracer_call_stack_depth]._rtol_op)
                {
                    case DBG_MQL_API_TRACING_FUNCTION_ID_ALERT:                 _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "Alert";         break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_COMMENT:               _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "Comment";       break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_PRINT:                 _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "Print";         break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_PRINTF:                _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "printf";        break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_PRINTFORMAT_ERROR:
                    case DBG_MQL_API_TRACING_FUNCTION_ID_PRINTFORMAT:           _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "PrintFormat";   break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_STRINGFORMAT:          _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "StringFormat";  break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_ACOS:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "acos";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_ASIN:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "asin";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_ATAN:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "atan";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_CEIL:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "ceil";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_COS:                   _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "cos";           break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_EXP:                   _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "exp";           break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_FABS:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "fabs";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_FLOOR:                 _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "floor";         break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_FMAX:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "fmax";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_FMIN:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "fmin";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_FMOD:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "fmod";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_LOG:                   _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "log";           break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_LOG10:                 _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "log10";         break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_POW:                   _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "pow";           break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_RAND:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "rand";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_ROUND:                 _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "round";         break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_SIN:                   _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "sin";           break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_SQRT:                  _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "sqrt";          break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_SRAND:                 _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "srand";         break;
                    case DBG_MQL_API_TRACING_FUNCTION_ID_TAN:                   _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = "tan";           break;
                    default:
                        {
                            string tmp[];
                            const int cnt = StringSplit(_mqlapi_func, 0x005F, tmp);
                            _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname = (cnt > 0) ? tmp[cnt - 1] : _global_trace_data[tracer_call_stack_depth]._mqlapi_funcname;
                        }
                }

                // Clear error code
                //::ResetLastError();

                // Return
                return(::GetPointer(this));
            };
    };


#ifndef __MQL4__
    };

#endif
//
/////////////////////////////////////////////////////////////////////////////////////////////////////



//
// END Debugging support
//*********************************************************************************************************************************************************/
#endif // LIB_DBG_DEBUG_MQLAPI_TRACER_MQH_INCLUDED
